/** \file hashtable.c
  * \brief Implements a hashtable with quadratic probing
  *
  * <PRE> Author: Anand Patwardhan
  * email: anand.patwardhan@umbc.edu
  * Date : 30 April 2004
  *
  * Implements a hashtable with quadratic probing, TCP sequence nos.
  * are used as keys for hashing packets, the hashtable itself does
  * not contain the raw packets, but pointer to raw packets.
  * TCP sequence nos. were chosen to be the keys since we intend to
  * perform intrusion detection on forwarded packets amongst other
  * things. If the HT is full, no more packets can be watched, this
  * however can be controlled by using an appropriate timeout period
  * for the timer which will flush packets deeming them to be dropped
  * after the timeout.
  * 
  * The functions makeEntry and PerformID are the HT equivalent of
  * insert and remove.
  *
  * The size for the hashtable should be a sufficiently large prime
  * no. to minimize collisions and reducing the chances of overflow
  *
  * Descriptions of each of the functions can be found in hashtable.h.
  *
  *****************************************************************************
  * The SNOOP program is an intrusion detection mechanism to detect local
  * intrusions in a Mobile Ad Hoc Network.
  *
  *  Copyright(C) 2005 University of Maryland, Baltimore County (UMBC)  
  *  E-mail: anand.patwardhan@umbc.edu
  *
  *  eBiquity Research Group
  *  University of Maryland, Baltimore County
  *  1000 Hilltop Circle, Baltimore, MD 21250, USA.
  *
  *  http://research.ebiquity.org
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  as published by the Free Software Foundation; either version 2
  *  of the License, or (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *******************************************************************************
  *  </PRE>
  */


#ifndef _HASHTABLE_C
#define _HASHTABLE_C

#include "common.h"
#include "hashtable.h"

hashtable* createHashtable(unsigned int size){
    unsigned int i;
    hashtable * ht;
    
    if(size<MINSIZE){
	printf("HT size should be at least %d",MINSIZE);
	exit(0);
    }
    
    ht = (hashtable *) malloc(sizeof(hashtable));
    
    if(NULL==ht){
	printf("Couldn't allocate memory for HT. Quitting ...\n");
	exit(0);
    }
    
    ht->table = (ht_bucket *) malloc(sizeof(ht_bucket)*size);
    
    if(NULL==ht->table){
	printf("Couldn't allocate memory for HT. Quitting ...\n");
	exit(0);
    }
    
    ht->size = size;
    ht->count = 0;
    
    for(i=0;i<size;i++){
	ht_bucket *b = &(const ht_bucket) ht->table[i];
	b->packet = (u_char *)0;
	b->state = INVALID;
	b->timestamp = 0;
    }

    return ht;
}

unsigned int hash(u_int32_t key, u_int32_t size){
    return key%size;
}

unsigned int locate(hashtable *ht, u_char * raw){
    int current, collisions, probes, emptyslot=-1;
    u_int32_t seqno;
    struct tcphdr* tcp_hdr;
    ht_bucket* htb;
    int etherlen = sizeof(struct ether_header);
    collisions = 0;
    probes = 0;

    
    tcp_hdr = (const struct tcphdr*) (raw + sizeof(struct ether_header) + sizeof(struct ip6_hdr));
    seqno = ntohl(tcp_hdr->seq);

    /* use the sequence no. from the TCP packet
     * since it is unlikely to be the same for two
     * different tcp streams at the same time
     * however will make it very fast when looking for that packet,
     * clearing the same entry when the node forwards it
     */
   
    current = hash(seqno, ht->size);

    /* however it is possible that something else is
     * occupying this space
     */
    
    htb = &(const ht_bucket) ht->table[current];

    /* if entry is valid compare  */
    while( probes <= ht->size &&
	    INVALID != htb->state){
	/*If bucket is valid, compare contents */
	if((VALID == htb->state)||(LATE == htb->state)){
	    // printf("Comparing : \n");
	    // hex_dump_packet(htb->packet);
	    // hex_dump_packet(raw);
	    // hop limit in ipv6 header changes by one !! need to ignore hop limit field
	    // should differ by one !
	    if( (0 == memcmp(htb->packet+etherlen, raw+etherlen, 7))
			    &&
		(0 == memcmp(htb->packet+etherlen+8, raw+etherlen+8, sizeof(struct ip6_hdr) + sizeof(struct tcphdr)-8))
			    ){
		// identical packet was found
		return current;
	    }
	} else if(DELETED == htb->state){
		// could have been marked for deletion by timeout
	        // free mem. if not already freed
		if(NULL!=htb->packet){
			free(htb->packet);
			htb->packet = NULL;
		}
		// mark this position as the first empty slot
		if(-1!=emptyslot)
			emptyslot=current;
	}
		
	++probes;
	current += 2 * ++collisions - 1;
        if( current >= ht->size )
	    current -= ht->size;

	htb = &(const ht_bucket) ht->table[current];
	//printf("FIND: probing, position=%d, state is : %d\n", current, htb->state);
    }
     if(probes > ht->size)
	 return emptyslot; /*if HT is full, emptyslot will be -1*/
     return current;
}


int makeEntry(hashtable * ht, u_char * raw, clock_t time){
    unsigned int pos = locate(ht, raw);
    ht_bucket * htb;

    if(pos!=-1){
	htb = &(ht_bucket) ht->table[pos];
        if((VALID==htb->state)||(LATE==htb->state)){
	    /*such a packet already exists,
	      so this must be a retransmission
	      by the same host, update timestamp
	      shouldnt we replace the old packet,
	      something might have changed this time ?!?!
	     */
	    
	    htb->timestamp = time;
	} else{ // either INVALID or DELETED
	    htb->packet = raw;
	    htb->timestamp = time;
	    htb->state = VALID;
	    (ht->count)++;
	  //  printf("Entry made in hashtable, SIZE = %d\n", ht->count);
	}
    }else{
	printf("Hash table full, cannot snoop ...\n");
	return 0;
    }

    return 1;
       
}

int performID(hashtable * ht, u_char * raw, clock_t time){
    unsigned int pos = locate(ht, raw);
    ht_bucket * htb;

    if(pos!=-1){
//	dump_packet(raw);
	htb = &(ht_bucket) ht->table[pos];
        if((VALID==htb->state)||(LATE==htb->state)){ 
        	// locate will return an empty slot position
	        // if no matching packet found, remove entry only if VALID
	        // such a packet already exists,
	        // so this must be the forwarded packet
	     
	    free(htb->packet);
	    htb->packet = NULL;
	    htb->timestamp = time;
	    htb->state = DELETED;
	    (ht->count)--;
	  // entry removed from HT
	}
    }else{
	printf("No matching packet found in hashtable\n");
	return 0;
    }

    //print(ht);

    return 1;
       
}


void print(hashtable *ht){
    int i;
    ht_bucket* htb;
    printf("*****    Hash Table contents    *********\n");
    for(i=0;i<ht->size;i++){
	htb = &(const ht_bucket) ht->table[i];
	if(VALID == htb->state){
	    printf("(V) [%u]:", htb->timestamp);
	    dump_packet(htb->packet);
	}
	if(LATE == htb->state){
	    printf("(L) [%u]:", htb->timestamp);
	    dump_packet(htb->packet);
	}

    }
    printf("*****************************************\n");
    
}

void dump_packet(u_char * packet){

    struct ip6_hdr * ip6;
    struct tcphdr * tcp;
    u_int32_t seqno;
    char *src;
    char ntop_buf[INET6_ADDRSTRLEN];
    u_int16_t sport, dport;
     
    ip6 = (const struct ip6_hdr*) (packet+ sizeof(struct ether_header));
    tcp = (const struct tcphdr*) (packet + sizeof(struct ether_header) + sizeof(struct ip6_hdr) );
    
    seqno = ntohl(tcp->seq);
    sport = ntohs(tcp->source);
    dport = ntohs(tcp->dest);

    src = inet_ntop(AF_INET6, &ip6->ip6_src, ntop_buf, sizeof(ntop_buf));
    printf("%s.%d >> ", src, sport);
    src = inet_ntop(AF_INET6, &ip6->ip6_dst, ntop_buf, sizeof(ntop_buf));
    printf("%s.%d %u\n", src, dport, seqno);
    
}

void hex_dump_packet(u_char * packet){
	int i,len;
	u_int8_t * raw = (const u_int8_t *) packet;
	for(i=0;i< (sizeof(struct ether_header) + sizeof(struct ip6_hdr) + sizeof(struct tcphdr));i++)
		printf("%x",raw[i]);
	printf("\n");
	
}



#endif
