/** \file hashtable.h
  * \brief Implements a hashtable with quadratic probing
  *
  * <PRE> Author: Anand Patwardhan
  * email: anand.patwardhan@umbc.edu
  * Date : 30 April 2004
  *
  * Implements a hashtable with quadratic probing, TCP sequence nos.
  * are used as keys for hashing packets, the hashtable itself does
  * not contain the raw packets, but pointer to raw packets.
  * TCP sequence nos. were chosen to be the keys since we intend to
  * perform intrusion detection on forwarded packets amongst other
  * things. If the HT is full, no more packets can be watched, this
  * however can be controlled by using an appropriate timeout period
  * for the timer which will flush packets deeming them to be dropped
  * after the timeout.
  * 
  * The functions makeEntry and PerformID are the HT equivalent of
  * insert and remove.
  *
  * The size for the hashtable should be a sufficiently large prime
  * no. to minimize collisions and reducing the chances of overflow
  *
  *****************************************************************************
  * The SNOOP program is an intrusion detection mechanism to detect local
  * intrusions in a Mobile Ad Hoc Network.
  *
  *  Copyright (C) 2005 University of Maryland, Baltimore County (UMBC) 
  *  E-mail: anand.patwardhan@umbc.edu
  *
  *  eBiquity Research Group
  *  University of Maryland, Baltimore County
  *  1000 Hilltop Circle, Baltimore, MD 21250, USA.
  *
  *  http://research.ebiquity.org
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  as published by the Free Software Foundation; either version 2
  *  of the License, or (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *******************************************************************************
  *  </PRE>
  */

#ifndef _HASHTABLE_H_
#define _HASHTABLE_H_

#include "common.h"

#define MINSIZE 11
#define HTSIZE  1279  // a large prime number, has to be greater than MINSIZE

//! Enumeration of states of a HT entry
typedef enum {
   VALID	= 1,
   INVALID	= 2,
   DELETED	= 3,
   LATE		= 4 
} ht_state;

//! Structure for a HT bucket
typedef struct {
    
   /// Pointer to raw packet
   u_char *packet;
   
   /// Whether the entry is valid, invalid, deleted or late
   ht_state state;
   
   /// timestamp when received
   time_t timestamp;
   
} ht_bucket;

//! Structure for a HT
typedef struct {
    
    /// Table of pointers to buckets
    ht_bucket *table;         
    
    ///  Maximum size 
    unsigned int size;
    
    /// how many entries
    unsigned int count;
    
} hashtable;


//! Create a hashtable of specified size
/*! \param size Size of the hashtable
 *  \return address of allocated HT
 */
hashtable* createHashtable(unsigned int size);

//! Add a new raw packet to the hashtable (inserts new entry into HT)
/*! Uses quadratic probing to make the new entry, uses TCP sequence
 *  number. Involves parsing the TCP header to get the sequence no.
 *  Thus currently only TCP packets can be watched, though any valid
 *  sized packet could still be entered, but not recommended. In the case
 *  of TCP packets, seq. nos. are ideal candidates for keys in the HT,
 *  especially when searching the HT to match an identical packet.
 *  \param ht Hashtable to make entry in
 *  \param raw Raw packet
 *  \param time Timestamp when packet was received
 */
int makeEntry(hashtable* ht, u_char* raw, clock_t time);

//! Performs comparison to see if a forwarded packet matches a received packet
/*! In the comparison, the hop limit field is ignored. The hop limit should be
 *  expected to decrease by one, if not, should be considered as a malicious
 *  modification, though not currently done in this code.
 * \param ht Specifies the hashtable
 * \param raw The raw packet contents
 * \param time The time when this packet was received
 */
int performID(hashtable* ht, u_char* raw, clock_t time);

//! Prints current contents of hashtable
/*! For debugging purposes, prints current contents of HT
 * \param ht Hashtable to print
 */
void print(hashtable* ht);

//! Prints the contents of a raw packet in human readable form
/*! \param packet The raw packet to be printed
 */
void dump_packet(u_char* packet);

//! Prints the hex contents of a raw packet
/*! \param packet The raw packet to be printed in hexadecimal format
 */
void hex_dump_packet(u_char* packet);

#endif
