/** \file snoop.c
  * \brief Uses the pcap library, to listen on any available network interface
  * in promiscuous mode. Raw packets are captured and further processed
  * for filtering, IPv6 packets are of interest. TCP streams over IPv6
  * are monitored. AODV6 packets are in UDP datagrams, again in IPv6
  * packets. AODVD hello messages are used to populate the neighbor table.
  * The IDS itself is independent of the routing protocol is in use. In this
  * instance we used SecAODV, so neighbors are discovered using AODV hello
  * messages.
  *
  * See snoop.h documentation for function descriptions.
  *
  * <PRE> Author: Anand Patwardhan
  * email: anand.patwardhan@umbc.edu
  * Date : 30 April 2004
  *
  *****************************************************************************
  * The SNOOP program is an intrusion detection mechanism to detect local
  * intrusions in a Mobile Ad Hoc Network.
  *
  *  Copyright(C) 2005 University of Maryland, Baltimore County (UMBC)  
  *  E-mail: anand.patwardhan@umbc.edu
  *
  *  eBiquity Research Group
  *  University of Maryland, Baltimore County
  *  1000 Hilltop Circle, Baltimore, MD 21250, USA.
  *
  *  http://research.ebiquity.org
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  as published by the Free Software Foundation; either version 2
  *  of the License, or (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *******************************************************************************
  *
  * We referred to Tim Carstens' example code and Martin Casado's
  * tutorial. We also looked at the ipgrab source code and tcpdump source.
  * Source code seems to be the only documentation on pcap.
  *  </PRE>
  */


#ifndef _SNOOP_C
#define _SNOOP_C

#include "common.h"
#include "hashtable.h"
#include "snoop.h"

hashtable * ht;
neighbor neighbors[2*MAXNEIGHBORS]; 
int received, forwarded;
FILE *logfile;

void SIGALRM_handler(int sig){
    printf("<<<<<<<< SIGARLM >>>>>>>>>>\n");
    signal(sig, SIGALRM_handler);
    printf("Hashtable size was : %d\n", ht->count);
    // make a pass through the hashtable and upcount
    timer_update_state();
    printf("Hashtable size now, is : %d\n", ht->count);
    printf("<<<<<<<<<<<<<>>>>>>>>>>>>>>\n");
    print_neighbors();
    log_intrusions();

    alarm(ALARM_TIMEOUT);
}


void timer_update_state(void){
    int i;
    ht_bucket* htb;
    struct ether_header* ether_hdr;
    struct ether_addr* ether_dst;
    for(i=0;i<ht->size;i++){
	htb = &(const ht_bucket) ht->table[i];
	switch(htb->state){
		case VALID:
			htb->state = LATE;
			break;
		case LATE:
			// already late, so assume dropped by node
			// upcount neighbor's dropped packets
    			ether_hdr = (struct ether_header*) htb->packet;
			ether_dst = (struct ether_addr*)(ether_hdr->ether_dhost);
			incr_dropcount(ether_dst);
			htb->state = DELETED; // mark for deletion
			(ht->count)--;
			break;
		case INVALID:
		case DELETED:
		default:
			// do nothing
	}
    }

}



/* looking at ethernet headers */
void pkt_callback(u_char *args,const struct pcap_pkthdr* pkthdr,const u_char* packet){
    u_int16_t type = handle_ethernet(args,pkthdr,packet);
    if(type == ETHERTYPE_IPV6){
 	/* handle IPv6 packet */
	handle_IPv6(args, pkthdr, packet);
    }
}

u_int16_t
handle_ethernet(u_char *args,const struct pcap_pkthdr* pkthdr,const u_char* packet){
    u_int caplen = pkthdr->caplen;
    u_int length = pkthdr->len;
    struct ether_header *eptr;  /* net/ethernet.h */
    u_short ether_type;

    if (caplen < ETHER_HDRLEN)
    {
        fprintf(stdout,"Incomplete ethernet header info\n");
        return -1;
    }

    eptr = (struct ether_header *) packet;
    ether_type = ntohs(eptr->ether_type);

    return ether_type;
}

int
handle_IPv6(u_char *args,const struct pcap_pkthdr* pkthdr,const u_char* packet){
  struct ip6_hdr *ip6;   // from netinet/ip6.h
  struct ether_header *ether;
  u_int length = pkthdr->len;  // length of packet from ethernet header
  u_int caplen = pkthdr->caplen;
  u_int hlen, off, version;
  int len,i;
  char *src;
  char ntop_buf[INET6_ADDRSTRLEN];
  u_int8_t nexthdr;
  const struct udphdr* udp_hdr;	// from netinet/udp.h
  const struct tcphdr* tcp_hdr; // from netinet/tcp.h
  u_int16_t sport, dport, udplen, udpchecksum;
  u_int8_t rawpacket[1500];
  u_char* packet_hdrs;
  u_int8_t aodv_type;
  u_int32_t seqno;
  neighbor n_pair;

  // skip past the ethernet header to the ip6 header

  ether = (const struct ether_header*) packet;

  ip6 = (const struct ip6_hdr *) (packet + sizeof(struct ether_header));
  
  length-=sizeof(struct ether_header);

  // check if packet is truncated

  if(length < sizeof(struct ip6_hdr)){
	  printf("Truncated ip6 packet %d\n",length);
	  return -1;
  }

  len = ntohs(ip6->ip6_plen);

  src = inet_ntop(AF_INET6, &ip6->ip6_src, ntop_buf, sizeof(ntop_buf));
  src = inet_ntop(AF_INET6, &ip6->ip6_dst, ntop_buf, sizeof(ntop_buf));

  memcpy(rawpacket, ip6, len);
  packet_hdrs = (u_char *) malloc(sizeof(struct ether_header) + sizeof(struct ip6_hdr) + sizeof(struct tcphdr));
  if(NULL==packet_hdrs){
      printf("Couldn't allocate mem. for packet, ignoring this one !\n");
  }
  else{
    memcpy(packet_hdrs, packet, sizeof(struct ether_header) + sizeof(struct ip6_hdr) + sizeof(struct tcphdr));
  }

  nexthdr = ip6->ip6_nxt;

  switch(nexthdr){
	  case IP_PROTO_TCP:
		  tcp_hdr = (const struct tcphdr *) (rawpacket + sizeof(struct ip6_hdr));
		  sport = ntohs(tcp_hdr->source);
		  dport = ntohs(tcp_hdr->dest);
		  seqno = ntohl(tcp_hdr->seq);

		  // check if destip == destether   => makeEntry
		  // and check if srcip == srcether => performID

		  memcpy(&n_pair.src_ether,&ether->ether_shost,6);
		  memcpy(&n_pair.src_ip6,&ip6->ip6_src,16);

		  if(-1==find_neighbor(&n_pair)){
		      // src ether != src ip6
		      printf("#############################\n");
		      printf("source(ether)!=source(ip6), FORWARDED = %d\n",++forwarded );
		      printf("#############################\n");

		      if(NULL!=packet_hdrs)
			  performID(ht, packet_hdrs,time(NULL));
		  }
		  
		  memcpy(&n_pair.src_ether,&ether->ether_dhost,6);
		  memcpy(&n_pair.src_ip6,&ip6->ip6_dst,16);

		  if(-1==find_neighbor(&n_pair)){
		      //dst ether != dest ip6
		      printf("#############################\n");
		      printf("dest(ether)!= dest(ip6), RECEIVED = %d\n", ++received);
		      printf("#############################\n");
		      // These are the packets that we want to snoop on
		      if(NULL!=packet_hdrs)
			  makeEntry(ht, packet_hdrs,time(NULL));
		  }
		  break;
		  
	  case IP_PROTO_UDP:
		  udp_hdr = (const struct udphdr *) (rawpacket + sizeof(struct ip6_hdr));
		  sport = ntohs(udp_hdr->source);
		  dport = ntohs(udp_hdr->dest);
		  udplen = ntohs(udp_hdr->len);
		  udpchecksum = ntohs(udp_hdr->check);

		  if((654==sport)&&(654==dport))
		      handle_AODV(packet);
		  // else non-AODV payload
		  break;
		  
	  case IP_PROTO_ICMPV6:
		  break;
		  
	  default:
		// other 
  }
  
  return 0;
}


int handle_AODV(const u_char* packet){
    u_int8_t type, flags, hops, prefixsize;
    u_int32_t reqid, dsn, osn, lifetime;
    struct ether_header * ether;
    struct ip6_hdr * ip6;
    struct aodv_rreq6_draft_01* rreq;
    struct aodv_rrep6_draft_01* rrep;
    neighbor aodv_pair, interface_pair;
    char * addr;
    char ntop_buf[INET6_ADDRSTRLEN];
    int i;
    union aodv* aodvpacket;
    


    aodvpacket = (const union aodv *)
	    (  packet
	       + sizeof(struct ether_header)
	       + sizeof(struct ip6_hdr)
	       + sizeof(struct udphdr) ) ;
    
    type = aodvpacket->rrep6.rrep_type;

    switch(type){
	case AODV6_RREQ:
	    rreq = (const struct aodv_rreq6_draft_01*) aodvpacket;
	    addr = inet_ntop(AF_INET6, &rreq->rreq_oa, ntop_buf, sizeof(ntop_buf));
	    addr = inet_ntop(AF_INET6, &rreq->rreq_da, ntop_buf, sizeof(ntop_buf));
	    break;
	case AODV6_RREP:

	    /* update neighbor table */
	    rrep = (const struct aodv_rrep6_draft_01*) aodvpacket;
	    addr = inet_ntop(AF_INET6, &rrep->rrep_oa, ntop_buf, sizeof(ntop_buf));
	    addr = inet_ntop(AF_INET6, &rrep->rrep_da, ntop_buf, sizeof(ntop_buf));

	    /*if hops = 0 and originator_ip = dest_ip add to neighbors*/

	    if((0==rrep->rrep_hops) && (0== memcmp(&rrep->rrep_oa,&rrep->rrep_da, sizeof(struct in6_addr)))){
		/*get the src ip from the ip6 header*/

		ether = (const struct ether_header*) packet;
		ip6 = (const struct ip6_hdr*) (packet+sizeof(struct ether_header));

		
		memcpy(&aodv_pair.src_ether,&ether->ether_shost,6);
		memcpy(&aodv_pair.src_ip6,&ip6->ip6_src, 16);
		memcpy(&interface_pair.src_ether,&ether->ether_shost,6);
		memcpy(&interface_pair.src_ip6,&rrep->rrep_oa, 16);

		if(-1==find_neighbor(&aodv_pair))
			add_neighbor(&aodv_pair);
		if(-1==find_neighbor(&interface_pair))
			add_neighbor(&interface_pair);
		
    
		
	    }
	    
	    break;
	default:
	    // other
      
    }

    return 0;
    
}

int find_neighbor(neighbor* pair){
    int i;
    for(i=0;i<2*MAXNEIGHBORS;i++){
	if((LIVE==neighbors[i].route_state)&&
		(0==memcmp(&neighbors[i],pair,22))){
	//	print_neighbors();
		return i;
	    }
    }
    return -1;
}

int add_neighbor(neighbor* pair){
    int i;
    for(i=0;i<2*MAXNEIGHBORS;i++){
	if(LIVE!=neighbors[i].route_state){
	    memcpy(&neighbors[i],pair,22);
	    neighbors[i].route_state = LIVE;
	    printf("Added a neighbor\n");
	   // print_neighbors();
	    return 1;
	}
    }
    return 0;
}

int incr_dropcount(struct ether_addr* ether_src){
	int i;
	for(i=0;i<2*MAXNEIGHBORS;i++){
		if(0==memcmp(&neighbors[i],ether_src,6)){
			neighbors[i].dropcount++;
		}
	}
	return 0;
}

void print_neighbors(void){
    int i,j;
    
    printf("\n====== Neighbor Table ======\n");
    printf("(dropcount, mac address, ip6 address)\n");
    for(i=0;i<2*MAXNEIGHBORS;i++)
	if(LIVE==neighbors[i].route_state){
	    printf("[%d] ", neighbors[i].dropcount);	
	    for(j=0;j<6;j++)
		printf("%x",neighbors[i].src_ether[j]);
	    printf(" = ");
	    for(j=0;j<16;j++)
		printf("%x",neighbors[i].src_ip6[j]);
	    printf("\n");
	}
    printf("============================\n");
}

void log_intrusions(void){
	// read the neighbor table, log non-zero dropcounts as intrusions
	// also reset dropcounts to zero
	int i;
	intruder badguy;
	const int n = sizeof(intruder);

	for(i=0;i<2*MAXNEIGHBORS;i++){
		if(LIVE==neighbors[i].route_state && neighbors[i].dropcount>0){
			memcpy(&badguy,&neighbors[i],22);
			badguy.when_detected = time(0);
			badguy.dropcount = neighbors[i].dropcount;
			// reset dropcount
			neighbors[i].dropcount = 0;

			//append to log
			// **** if error handling is attempted, doesnt write to file
			// just dont doubt it and it works :D
			//if(fwrite(&badguy, sizeof(intruder), 1, logfile)!=1)
			fwrite(&badguy, sizeof(intruder), 1, logfile);
			fflush(logfile);

			
		}// end if
	}// end for

	
}

int main(int argc,char **argv){
 
    char *dev; 
    char errbuf[PCAP_ERRBUF_SIZE];
    pcap_t* descr;
    struct bpf_program fp;      /* hold compiled program     */
    bpf_u_int32 maskp;          /* subnet mask               */
    bpf_u_int32 netp;           /* ip                        */
    u_char* args = NULL;
    int i;
    received = 0;
    forwarded = 0;


    if(argc < 2){ 
        fprintf(stdout,"Usage: %s numpackets \"options\"\n",argv[0]);
        return 0;
    }

    // open log file
    if((logfile = fopen("/tmp/suspects","w")) == NULL){
	    printf("Couldn't open log file ...\n");
	    exit(0);
    }

    dev = pcap_lookupdev(errbuf);
    if(dev == NULL)
    { printf("%s\n",errbuf); exit(1); }


    ht = createHashtable(HTSIZE); /* size is a suitably large prime no. */

    for(i=0;i<2*MAXNEIGHBORS;i++){
	neighbors[i].route_state = EXPIRED;
    }

    /* signal handling */

    if(SIG_ERR==signal(SIGALRM, SIGALRM_handler)){
	printf("SIGALRM handler couldnt be installed ...\n");
	exit(1);
    }

    alarm(ALARM_TIMEOUT);

    /* get the network address and mask of the device */
    pcap_lookupnet(dev,&netp,&maskp,errbuf);

    /* open device in promiscuous mode */
    
    descr = pcap_open_live(dev,1500,1,-1,errbuf);
    if(descr == NULL)
    { printf("pcap_open_live(): %s\n",errbuf); exit(1); }


    if(argc > 2)
    {
        if(pcap_compile(descr,&fp,argv[2],0,netp) == -1)
        { fprintf(stderr,"Error calling pcap_compile\n"); exit(1); }

        if(pcap_setfilter(descr,&fp) == -1)
        { fprintf(stderr,"Error setting filter\n"); exit(1); }
    }

    /* begin snooping */ 
    pcap_loop(descr,atoi(argv[1]),pkt_callback,args);

    fprintf(stdout,"\nDone...\n");
    print_neighbors();
    fclose(logfile);
	
    return 0;
}

#endif



