/** \file snoop.h
  * \brief Uses the pcap library, to listen on any available network interface
  * in promiscuous mode. Raw packets are captured and further processed
  * for filtering, IPv6 packets are of interest. TCP streams over IPv6
  * are monitored. AODV6 packets are in UDP datagrams, again in IPv6
  * packets. AODVD hello messages are used to populate the neighbor table
  *
  * <PRE> Author: Anand Patwardhan
  * email: anand.patwardhan@umbc.edu
  * Date : 30 April 2004
  *
  *****************************************************************************
  * The SNOOP program is an intrusion detection mechanism to detect local
  * intrusions in a Mobile Ad Hoc Network.
  *
  *  Copyright (C) 2005 University of Maryland, Baltimore County (UMBC) 
  *  E-mail: anand.patwardhan@umbc.edu
  *
  *  eBiquity Research Group
  *  University of Maryland, Baltimore County
  *  1000 Hilltop Circle, Baltimore, MD 21250, USA.
  *
  *  http://research.ebiquity.org
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  as published by the Free Software Foundation; either version 2
  *  of the License, or (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *******************************************************************************
  *
  * We referred to Tim Carstens' example code and Martin Casado's
  * tutorial. We also looked at the ipgrab source code and tcpdump source.
  * Source code seems to be the only documentation on pcap.
  *  </PRE>
  */


#ifndef _SNOOP_H
#define _SNOOP_H
#include "common.h"

//! MAXNEIGHBORS is used to limit the number of neigbors to watch for intrusions
/*! To be scalable, it is necessary to watch only a bounded number of neighbors
  * depending on the memory and cpu capacity of the device, or else it can be
  * quickly overwhelmed and effectiveness of the IDS will decrease.
  */
#define MAXNEIGHBORS 10

//! ALARM_TIMEOUT is the value in seconds for the timeout to examine the hashtable
/*! The hashtable is examined every ALARM_TIMEOUT seconds; a second chance algorithm
 *  is used to classify packet drops as intrusions. Basically if a packet is not
 *  retransmitted by the time of a second timeout, it is considered to be dropped
 *  and removed from the hashtable.
 */
#define ALARM_TIMEOUT 3 


//! Three states of a neighbor within the neighbor table
typedef enum {EXPIRED, LIVE, BAD} r_state;

//! The neighbor struct stores that mac address and corresponding IPv6 address of its neighbors
typedef struct{
    
    /// source ethernet address 
    u_int8_t src_ether[6];
    
    /// corresponding ip6 address 
    u_int8_t src_ip6[16];
    
    /// live, bad or expired
    r_state route_state;
    
    /// count of dropped packets since last timeout
    u_int32_t dropcount;
    
    /// when this route will expire 
    clock_t expiry;
    
} neighbor;

//! The intruder struct holds the mac, ipv6 addresses, the time when it was classified as intrusive and the dropcount for that period
typedef struct{
    
	/// ethernet address
	u_int8_t ether_addr[6];
	
	/// ip6 address
	u_int8_t ip6_addr[16];
	
	/// end of sampling period is recorded
	time_t   when_detected;
	
	/// right now there is just one type, need (type,count) for each type of attack
	u_int32_t dropcount;    
	
} intruder;

//! Alarm handler, updates drop counts
/*! Calls the timer_update_state function
 */
void SIGALRM_handler(int);

//! Examines the hashtable and clears up packets, updates drop counts
/*! Will remove entries from the hashtable that are old, gives newer ones
 * a second chance, if entries are not cleared by a second timeout, the packet
 * is assumed to have been dropped.
 */
void timer_update_state(void);

//! The callback function for the packet capture
/*! The callback function opens the device in promiscuous mode and listens for
  * packets. The raw packets captured (1500 bytes) are then passed on for further
  * filtering. We deal with only IPv6 packets, others are ignored.
 */
void pkt_callback(u_char *, const struct pcap_pkthdr *, const u_char*);

//! Returns the type of packet contained within the ethernet frame
/*! We are interested only in IPv6, but can be easily modified to handle other
  * protocols
  */
u_int16_t handle_ethernet(u_char *, const struct pcap_pkthdr *, const u_char*);

//! Examines IPv6 packets for AODV6 and TCP payloads
/*! Once IPv6 packets are filtered out, further filtering of AODV6 and TCP
  * protocols is done here. For AODV6 we watch for "Hello" messages i.e. special
  * RREP messages, to identify neigbhors and populate the neigbor table.
  */
int handle_IPv6(u_char *, const struct pcap_pkthdr *, const u_char*);

//! Examines and handles AODV packets
/*! Looks for RREP messages to build the neigbor tables. We can potentially
 *  remove entries from the neigbor table if hello messages are missed.
 */
int handle_AODV(const u_char*);

//! Helper function that returns the index of the provided neighbor in the neighbor table.
/*! Provided a neighbor instance, looks up and returns the index of the neighbor
  * in the neighbor table.
  */
int find_neighbor(neighbor *);

//! Helper function that adds a neighbor to the neighbor table
int add_neighbor(neighbor *);

//! Helper function that prints contents of neighbor table
void print_neighbors(void);

//! Increments the dropcount of the entry in the neigbor table corresponding to the provided mac address
int incr_dropcount(struct ether_addr*);

//! Reads the neighbor table, logs non-zero dropcount entries as potential intrusions.
/*! This function reads the neighbor table, logs non-zero entries to the suspects file
  * and resets dropcounts for the next time period.
  */
void log_intrusions(void);

#endif
